def dict_to_html(d, indent=0):
    from IPython.display import HTML

    html = ""
    for key, value in d.items():
        if isinstance(value, dict):
            html += f"<div style='margin-left: {indent}em;'><details><summary>{key}</summary>{dict_to_html(value, indent + 1)}</details></div>"
        else:
            html += f"<div style='margin-left: {indent}em;'><strong>{key}:</strong> {value}</div>"
    return html



import os
import glob
import json
import pkgutil
import pathlib
import inspect

import functools
import importlib

import toolviper.utils.logger
import toolviper.utils.console as console

from toolviper.utils.protego import Protego

from typing import Callable, Any, Union, NoReturn, Dict, List, Optional, Tuple
from types import ModuleType


def is_notebook() -> bool:
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True
        else:
            raise ImportError

    except ImportError:
        return False


def validate(
    config_dir: str = None,
    custom_checker: Callable = None,
    add_data_type: Any = None,
    external_logger: Callable = None,
):
    def function_wrapper(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            meta_data = {"function": None, "module": None}
            arguments = inspect.getcallargs(function, *args, **kwargs)

            meta_data["function"] = function.__name__
            meta_data["module"] = function.__module__

            # If this is a class method, drop the self entry.
            if "self" in list(arguments.keys()):
                class_name = args[0].__class__.__name__
                meta_data["function"] = ".".join((class_name, function.__name__))
                del arguments["self"]

            verify(
                function=function,
                args=arguments,
                config_dir=config_dir,
                custom_checker=custom_checker,
                add_data_type=add_data_type,
                external_logger=external_logger,
                meta_data=meta_data,
            )

            return function(*args, **kwargs)

        return wrapper

    return function_wrapper


# DEPRECATED
def _get_path(function: Callable) -> str:
    module = inspect.getmodule(function)
    module_path = inspect.getfile(module).rstrip(".py")

    if "src" in module_path:
        # This represents a local developer install

        base_module_path = module_path.split("src/")[0]
        return base_module_path

    else:
        # Here we hope that we can find the package in site-packages and it is unique
        # otherwise the user should provide the configuration path in the decorator.

        base_module_path = module_path.split("site-packages/")[0]
        return str(pathlib.Path(base_module_path).joinpath("site-packages/"))


def get_path(function: Callable) -> tuple[str, str]:
    module = inspect.getmodule(function)
    module_path = inspect.getfile(module).removesuffix(".py")

    # Determine whether this is a developer install or a site install
    tag = "src" if "src" in module_path else "site-packages"

    try:
        # The base directory should be to the left of the tag
        base_module_path = module_path.split(f"{tag}/")[0]

        # Split the module path up and determine what the package name and location is.
        split_path = module_path.split("/")
        index = split_path.index(tag) + 1
        package_name = split_path[index]

        # Build the full package path
        base_module_path = pathlib.Path(base_module_path).joinpath(
            f"{tag}/{package_name}"
        )

        return str(base_module_path), module_path

    except ValueError:
        import importlib

        toolviper.utils.logger.debug(
            f"module {module.__name__} has non-standard install path ..."
        )

        # Need to get the base package name
        package = module.__name__.split(".")[0]

        # IF we load the base pacakge we can find the base path
        package_path = importlib.import_module(package).__file__
        base_module_path = pathlib.Path(package_path).parent

        return str(base_module_path), module_path


def config_search(root: str = "/", module_name=None) -> Union[None, str]:
    colorize = console.Colorize()

    if root == "/":
        toolviper.utils.logger.warning("File search from root could take some time ...")

    toolviper.utils.logger.debug(
        "Searching {} for configuration file, please wait ...".format(
            colorize.blue(root)
        )
    )

    for file in glob.glob("{root}/**".format(root=root), recursive=True):
        if module_name + ".param.json" in file:
            basename = os.path.dirname(file)
            return basename

    return None


def set_config_directory(path: str, create: bool = False) -> NoReturn:
    colorize = console.Colorize()
    if pathlib.Path(path).exists():
        toolviper.utils.logger.info(
            "Setting configuration directory to [{path}]".format(
                path=colorize.blue(path)
            )
        )
        os.environ["PARAMETER_CONFIG_PATH"] = path
    else:
        toolviper.utils.logger.info(
            "The configuration directory [{path}] does not currently exist.".format(
                path=colorize.blue(path)
            )
        )
        if create:
            toolviper.utils.logger.info(
                "Creating empty configuration directory: {path}".format(
                    path=colorize.blue(path)
                )
            )
            pathlib.Path(path).mkdir()


def verify_configuration(path: str, module: ModuleType) -> List[str]:
    modules = []
    for file in glob.glob("{path}/*.param.json".format(path=path), recursive=True):
        if file.endswith(".param.json"):
            modules.append(os.path.basename(file).strip(".param.json"))

    package_path = os.path.dirname(module.__file__)
    package_modules = [name for _, name, _ in pkgutil.iter_modules([package_path])]

    not_found = []
    for module in package_modules:
        if module not in modules:
            not_found.append(module)

    # coverage = (len(package_modules) - len(not_found)) / len(package_modules)

    return package_modules


def verify(
    function: Callable,
    args: Dict,
    meta_data: Dict[str, Union[Optional[str], Any]],
    config_dir: str = None,
    add_data_type: Any = None,
    custom_checker: Callable = None,
    external_logger: Callable = None,
) -> NoReturn:
    colorize = console.Colorize()
    function_name, module_name = meta_data.values()

    # The module call gives the full module chain, but we only want the last
    # module, ex. astrohack.extract_holog would be returned, but we only want
    # extract_holog. This should generally work.

    module_name = module_name.split(".")[-1]

    if external_logger is None:
        logger = toolviper.utils.logger.get_logger()

    else:
        logger = external_logger

    toolviper.utils.logger.debug(
        "Checking parameter values for {module}.{function}".format(
            function=colorize.blue(function_name), module=colorize.blue(module_name)
        )
    )

    path = None

    package_path, module_path = get_path(function)
    logger.info(f"Module path: {colorize.blue(package_path)}")

    # First we need to find the parameter configuration files
    if (
        pathlib.Path(package_path)
        .joinpath("config")
        .joinpath(f"{module_name}.param.json")
        .exists()
    ):
        logger.debug(
            f"Found configuration for {module_name}.{function_name} in: {colorize.blue(package_path)}"
        )
        path = str(pathlib.Path(package_path).joinpath("config"))

    # User specified configuration directory take precedent
    if config_dir is not None:
        if pathlib.Path(config_dir).joinpath(f"{module_name}.param.json").exists():
            logger.debug(
                f"Setting configuration directory to user provided [{config_dir}]"
            )
            path = config_dir

        else:
            logger.warning(
                "User provided configuration directory does not exist. Searching for parameter files ..."
            )

    # If we have been delt only failure at this point, then we try one last ditch effort. Search the package directory!
    if not path:
        logger.debug(
            f"Couldn't determine parameter configuration directory, doing a depth search of {package_path}"
        )
        path = config_search(root=package_path, module_name=module_name)

        if not path:
            logger.error(
                f"Cannot find parameter configuration directory for {function_name}"
            )
            raise FileNotFoundError

    # Define parameter file name
    parameter_file = module_name + ".param.json"

    logger.debug(
        f"Parameter configuration file: {pathlib.Path(path).joinpath(parameter_file)}"
    )

    with open(pathlib.Path(path).joinpath(parameter_file)) as json_file:
        schema = json.load(json_file)

    if function_name not in schema.keys():
        logger.error(
            "{function} not_found in parameter configuration files.".format(
                function=colorize.format(function_name, color="red", bold=True)
            )
        )

        raise KeyError

    # Here is where the parameter checking is done.
    # First instantiate the validator.
    validator = Protego()

    # Register any additional data types that might be in the configuration file.
    if add_data_type is not None:
        validator.register_data_type(add_data_type)

    # Set up the schema to validate against.
    validator.schema = schema[function_name]

    # If a custom unit custom checker is needed, instantiate the
    # void function in the validator class. In the schema this
    # is used with "check allowed with".
    if custom_checker is not None:
        validator.custom_allowed_function = custom_checker

    assert validator.validate(args), logger.error(validator.errors)



import os
import shutil
import requests
import zipfile
import json
import psutil
import pathlib

import toolviper
import concurrent.futures

import toolviper.utils.logger as logger

from typing import NoReturn, Union
import toolviper.utils.console as console

colorize = console.Colorize()


def version():
    # Load the file dropbox file meta data.
    meta_data_path = pathlib.Path(__file__).parent.joinpath(
        ".dropbox/file.download.json"
    )

    # Verify that the download metadata exists and updates if not.
    _verify_metadata_file()

    with open(meta_data_path) as json_file:
        file_meta_data = json.load(json_file)

        logger.info(f'{file_meta_data["version"]}')


def dropbox(
    file: Union[str, list],
    folder: str = ".",
    threaded: bool = True,
    n_threads: Union[None, int] = None,
) -> NoReturn:
    """
        Download tool for data stored externally.
    Parameters
    ----------
    file : str
        Filename as stored on external source.
    folder : str
        Destination folder.
    threaded : bool
        File metadata download type.
    n_threads : int
        Number of threads to use.

    Returns
    -------
        No return
    """

    update()

    if not pathlib.Path(folder).resolve().exists():
        toolviper.utils.logger.info(
            f"Creating path:{colorize.blue(str(pathlib.Path(folder).resolve()))}"
        )
        pathlib.Path(folder).resolve().mkdir()

    if threaded is False:
        _download(file=file, folder=folder)

    else:

        if not isinstance(file, list):
            file = [file]

        if n_threads is None:
            n_threads = _get_usable_threads(len(file))

        logger.debug(f"Initializing downloader with {n_threads} threads.")

        _print_file_queue(file)

        with concurrent.futures.ThreadPoolExecutor(max_workers=n_threads) as executor:
            for _file in file:
                executor.submit(_download, _file, folder)


def list_files():
    from rich.table import Table
    from rich.console import Console

    console = Console()

    table = Table(show_header=True, show_lines=True)

    meta_data_path = pathlib.Path(__file__).parent.joinpath(
        ".dropbox/file.download.json"
    )

    # Verify that the download metadata exist and update if not.
    _verify_metadata_file()

    with open(meta_data_path) as json_file:
        file_meta_data = json.load(json_file)

        table.add_column("file", style="blue")
        table.add_column("dtype", style="green")
        table.add_column("telescope", style="green")
        table.add_column("size", style="green")
        table.add_column("mode", style="green")

        for filename in file_meta_data["metadata"].keys():
            values = [filename]

            for key, value in file_meta_data["metadata"][filename].items():
                if key in ["dtype", "telescope", "size", "mode"]:
                    values.append(value)

            table.add_row(*values)

    console.print(table)


def get_files():
    meta_data_path = pathlib.Path(__file__).parent.joinpath(
        ".dropbox/file.download.json"
    )

    # Verify that the download metadata exists and updates if not.
    _verify_metadata_file()

    with open(meta_data_path) as json_file:
        file_meta_data = json.load(json_file)

        return list(file_meta_data["metadata"].keys())


def update():
    meta_data_path = pathlib.Path(__file__).parent.joinpath(".dropbox")

    _makedir(str(pathlib.Path(__file__).parent), ".dropbox")

    file_meta_data = {
        "metadata": {
            "file.download.json": {
                "file": "file.download.json",
                "id": "1m53led1mchpdc4m3pv37",
                "rlkey": "enkp8m1hv437nu6p020owflrt&st=11psoc6n",
            }
        }
    }

    logger.info("Updating file metadata information ... ")

    # Download metadata without visual indicator bar.
    _get_from_dropbox(
        file="file.download.json",
        folder=str(meta_data_path),
        file_meta_data=file_meta_data,
        bar=False,
    )

    # assert meta_data_path.exists() is True, logger.error("Unable to retrieve download metadata.")


def _get_usable_threads(n_files: int) -> int:
    # Always leave a single thread resource
    available_threads = psutil.cpu_count(logical=True) - 1

    if available_threads >= n_files:
        return n_files

    return int(available_threads)


def _print_file_queue(files: list) -> NoReturn:
    from rich.table import Table
    from rich.console import Console
    from rich import box

    console = Console()
    table = Table(show_header=True, box=box.SIMPLE)

    table.add_column("Download List", justify="left")

    for file in files:
        table.add_row(f"[magenta]{file}[/magenta]")

    console.print(table)


def _is_notebook() -> bool:
    """
        Determines if code is running in  jupyter notebook.
    Returns
    -------
        bool

    """
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True
        else:
            raise ImportError

    except ImportError:
        return False


def _get_from_dropbox(file: str, folder: str, file_meta_data: dict, bar=True) -> None:
    fullname = file_meta_data["metadata"][file]["file"]
    id = file_meta_data["metadata"][file]["id"]
    rlkey = file_meta_data["metadata"][file]["rlkey"]

    url = "https://www.dropbox.com/scl/fi/{id}/{file}?rlkey={rlkey}".format(
        id=id, file=fullname, rlkey=rlkey
    )

    r = requests.get(url, stream=True, headers={"user-agent": "Wget/1.16 (linux-gnu)"})
    total = int(r.headers.get("content-length", 0))

    fullname = str(pathlib.Path(folder).joinpath(fullname))

    if _is_notebook():
        # This is due to a failure in python > 3.10 with tdqm fix this later
        # from tqdm.notebook import tqdm
        from tqdm import tqdm

    else:
        from tqdm import tqdm

    print(" ", end="", flush=True)

    # Is there a cleaner way to do this?
    if bar:
        with (
            open(fullname, "wb") as fd,
            tqdm(
                desc=fullname,
                total=total,
                unit="iB",
                unit_scale=True,
                unit_divisor=1024,
            ) as bar,
        ):
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    size = fd.write(chunk)
                    bar.update(size)

    else:
        with open(fullname, "wb") as fd:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    size = fd.write(chunk)


def _download(file: str, folder: str = ".") -> NoReturn:
    """
        Download tool for data stored on dropbox.
    Parameters
    ----------
    file : str
        Filename as stored on dropbox.
    folder : str
        Destination folder.

    Returns
    -------
        No return
    """

    # Load the file dropbox file meta data.
    meta_data_path = pathlib.Path(__file__).parent.joinpath(
        ".dropbox/file.download.json"
    )

    if meta_data_path.exists():
        with open(meta_data_path) as json_file:
            file_meta_data = json.load(json_file)

        full_file_path = pathlib.Path(folder).joinpath(file)

        if full_file_path.exists():
            logger.info("File exists: {file}".format(file=str(full_file_path)))
            return

        if file not in file_meta_data["metadata"].keys():
            logger.error(f"Requested file not found: {file}")
            logger.info(
                f"For a list of available files try using "
                f"{colorize.blue('toolviper.utils.data.list_files()')}."
            )

            return

    else:

        logger.wanring(
            f"Couldn't find file metadata locally in {colorize.blue(str(meta_data_path))}, trying to retrieve ..."
        )

        update()

        return

    fullname = file_meta_data["metadata"][file]["file"]
    fullname = str(pathlib.Path(folder).joinpath(fullname))

    _get_from_dropbox(file, folder, file_meta_data)

    if zipfile.is_zipfile(fullname):
        shutil.unpack_archive(filename=fullname, extract_dir=folder)

        # Let's clean up after ourselves
        os.remove(fullname)


def _makedir(path, folder):
    p = pathlib.Path(path).joinpath(folder)
    try:
        p.mkdir()
        logger.info(
            f"Creating path:{colorize.blue(str(pathlib.Path(folder).resolve()))}"
        )

    except FileExistsError:
        logger.warning(f"File exists: {colorize.blue(str(p.resolve()))}")

    except FileNotFoundError:
        logger.warning(
            f"One fo the parent directories cannot be found: {colorize.blue(str(p.resolve()))}"
        )


def _verify_metadata_file():
    meta_data_path = pathlib.Path(__file__).parent.joinpath(
        ".dropbox/file.download.json"
    )

    if not meta_data_path.exists():
        logger.warning(f"Couldn't find {colorize.blue(meta_data_path)}.")
        update()



from .download import dropbox
from .cloudflare import download, list_files, get_files, update, version

# from .cloudflare import download, version, list_files, get_files, update
__all__ = [s for s in dir() if not s.startswith("_")]



import os
import shutil
import requests
import zipfile
import json
import pathlib

import toolviper

import toolviper.utils.logger as logger
import toolviper.utils.console as console

from threading import Thread
from rich.progress import Progress

from typing import NoReturn, Union, Optional

colorize = console.Colorize()

PROGRESS_MAX_CHARACTERS = 28
MINIMUM_CHUNK_SIZE = 1024


def version() -> NoReturn:
    # Load the file dropbox file meta data.
    meta_data_path = pathlib.Path(__file__).parent.joinpath(
        ".cloudflare/file.download.json"
    )

    # Verify that the download metadata exists and updates if not.
    _verify_metadata_file()

    with open(meta_data_path) as json_file:
        file_meta_data = json.load(json_file)

        logger.info(f'{file_meta_data["version"]}')


def download(
    file: Union[str, list],
    folder: str = ".",
    overwrite: bool = False,
) -> NoReturn:
    """
        Download tool for data stored externally.
    Parameters
    ----------
    file : str
        Filename as stored on an external source.
    folder : str
        Destination folder.
    overwrite : bool
        Should file be overwritten.

    Returns
    -------
        No return
    """

    logger.info("Downloading from [cloudflare] ....")

    if not isinstance(file, list):
        file = [file]

    try:
        _print_file_queue(file)

    except Exception as e:
        logger.warning(f"There was a problem printing the file list... {e}")

    finally:
        if not pathlib.Path(folder).resolve().exists():
            toolviper.utils.logger.info(
                f"Creating path:{colorize.blue(str(pathlib.Path(folder).resolve()))}"
            )
            pathlib.Path(folder).resolve().mkdir()

    logger.debug(f"Initializing [cloudflare] downloader ...")

    meta_data_path = pathlib.Path(__file__).parent.joinpath(
        ".cloudflare/file.download.json"
    )

    tasks = []

    # Make a list of files that aren't available from cloudflare yet
    missing_files = []

    # Load the file dropbox file meta data.
    if not meta_data_path.exists():
        logger.warning(
            f"Couldn't find file metadata locally in {colorize.blue(str(meta_data_path))}"
        )

        toolviper.utils.data.update()

    with open(meta_data_path) as json_file:
        file_meta_data = json.load(json_file)

        # Build the task list
        for file_ in file:
            full_file_path = pathlib.Path(folder).joinpath(file_)

            if full_file_path.exists() and not overwrite:
                logger.info(f"File exists: {str(full_file_path)}")
                continue

            if file_ not in file_meta_data["metadata"].keys():
                logger.error(f"Requested file not found: {file_}")
                logger.info(
                    f"For a list of available files try using "
                    f"{colorize.blue('toolviper.utils.data.list_files()')}."
                )

                missing_files.append(file_)
                continue

            name_format = lambda string: (
                f"{string[:(PROGRESS_MAX_CHARACTERS - 4)]} ..."
                if len(string) > PROGRESS_MAX_CHARACTERS
                else string
            )

            tasks.append(
                {
                    "description": name_format(file_),
                    "metadata": file_meta_data["metadata"][file_],
                    "folder": folder,
                    "visible": True,
                    "size": int(file_meta_data["metadata"][file_]["size"]),
                }
            )

    threads = []
    progress = Progress()

    with progress:
        task_ids = [
            # progress.add_task(task["description"], total=task["size"])
            progress.add_task(task["description"])
            for task in tasks
            if len(tasks) > 0
        ]

        for i, task in enumerate(tasks):
            thread = Thread(target=worker, args=(progress, task_ids[i], task))
            thread.start()
            threads.append(thread)

        for thread in threads:
            thread.join()

    if len(missing_files) > 0:
        logger.info(f"Trying to retrieve missing files dropbox: {missing_files}")
        toolviper.utils.data.dropbox(file=missing_files, folder=folder)


def worker(progress: Progress, task_id: int, task: dict) -> NoReturn:
    """Simulate work being done in a thread"""

    fullname = task["metadata"]["file"]

    url = (
        f"http://downloadnrao.org/{task['metadata']['path']}/{task['metadata']['file']}"
    )

    r = requests.get(url, stream=True, headers={"user-agent": "Wget/1.16 (linux-gnu)"})
    total = int(r.headers.get("Content-Length", 0))

    if total == 0:
        total = task["size"]

    fullname = str(pathlib.Path(task["folder"]).joinpath(fullname))

    size = 0
    with open(fullname, "wb") as fd:

        for chunk in r.iter_content(chunk_size=MINIMUM_CHUNK_SIZE):
            if chunk:
                size += fd.write(chunk)
                progress.update(
                    task_id, completed=size, total=total, visible=task["visible"]
                )

    if zipfile.is_zipfile(fullname):
        shutil.unpack_archive(filename=fullname, extract_dir=task["folder"])

        # Let's clean up after ourselves
        os.remove(fullname)


def list_files() -> NoReturn:
    """
    List all files in cloudflare
    """

    from rich.table import Table
    from rich.console import Console

    console = Console()

    table = Table(show_header=True, show_lines=True)

    meta_data_path = pathlib.Path(__file__).parent.joinpath(
        ".cloudflare/file.download.json"
    )

    # Verify that the download metadata exist and update if not.
    _verify_metadata_file()

    with open(meta_data_path) as json_file:
        file_meta_data = json.load(json_file)

        table.add_column("file", style="blue")
        table.add_column("dtype", style="green")
        table.add_column("telescope", style="green")
        table.add_column("size", style="green")
        table.add_column("mode", style="green")

        for filename in file_meta_data["metadata"].keys():
            values = [filename]

            for key, value in file_meta_data["metadata"][filename].items():
                if key in ["dtype", "telescope", "size", "mode"]:
                    values.append(value)

            table.add_row(*values)

    console.print(table)


def get_files() -> NoReturn:
    """
    Get all files available in cloudflare manifest. This is retrieved from the local cloudflare
    metadata file.

    """
    meta_data_path = pathlib.Path(__file__).parent.joinpath(
        ".cloudflare/file.download.json"
    )

    # Verify that the download metadata exists and updates if not.
    _verify_metadata_file()

    with open(meta_data_path) as json_file:
        file_meta_data = json.load(json_file)

        return list(file_meta_data["metadata"].keys())


def update() -> NoReturn:
    """
    Update cloudflare manifest.
    """
    meta_data_path = pathlib.Path(__file__).parent.joinpath(".cloudflare")

    _makedir(str(pathlib.Path(__file__).parent), ".cloudflare")

    file_meta_data = {
        "file": "file.download.json",
        "path": "/",
        "dtype": "JSON",
        "telescope": "NA",
        "size": "12484",
        "mode": "NA",
    }

    tasks = {
        "description": "file.download.json",
        "metadata": file_meta_data,
        "folder": meta_data_path,
        "visible": False,
        "size": 12484,
    }

    logger.info("Updating file metadata information ... ")

    progress = Progress()
    task_id = progress.add_task(tasks["description"])

    with progress:
        worker(progress, task_id, tasks)

    if not meta_data_path.exists():
        logger.error("Unable to retrieve download metadata.")
        raise FileNotFoundError(
            "Download metadata file does not exist at the expected path."
        )


def get_file_size(path: str) -> Optional[dict]:
    """
    Get list file sizes in bytes for a given path. Only works for files; isn't recursive.
    """
    if not pathlib.Path(path).resolve().exists():
        logger.error(f"Path not found...: {path}")

        return None

    file_size_dict = {}

    for item in pathlib.Path(path).resolve().iterdir():
        if pathlib.Path(item).resolve().is_file():
            if item.name.endswith(".zip"):
                item_ = item.name.split(".zip")[0]

            else:
                item_ = item.name

            file_size_dict[item_] = os.path.getsize(pathlib.Path(item))

    return file_size_dict


def _print_file_queue(files: list) -> NoReturn:
    from rich.table import Table
    from rich.console import Console
    from rich import box

    console = Console()
    table = Table(show_header=True, box=box.SIMPLE)

    table.add_column("Download List", justify="left")

    for file in files:
        table.add_row(f"[magenta]{file}[/magenta]")

    console.print(table)


def _makedir(path, folder):
    p = pathlib.Path(path).joinpath(folder)
    try:
        p.mkdir()
        logger.info(
            f"Creating path:{colorize.blue(str(pathlib.Path(folder).resolve()))}"
        )

    except FileExistsError:
        logger.warning(f"File exists: {colorize.blue(str(p.resolve()))}")

    except FileNotFoundError:
        logger.warning(
            f"One fo the parent directories cannot be found: {colorize.blue(str(p.resolve()))}"
        )


def _verify_metadata_file():
    meta_data_path = pathlib.Path(__file__).parent.joinpath(
        ".cloudflare/file.download.json"
    )

    if not meta_data_path.exists():
        logger.warning(f"Couldn't find {colorize.blue(meta_data_path)}.")
        update()



import os
import re
import pathlib
import distributed

from toolviper.dask import menrva
from toolviper.dask.client import local_client


class TestToolViperMenerva:
    @classmethod
    def setup_class(cls):
        """setup any state specific to the execution of the given test class
        such as fetching test data"""
        pass

    @classmethod
    def teardown_class(cls):
        """teardown any state that was previously setup with a call to setup_class
        such as deleting test data"""
        # cls.client.shutdown()
        pass

    def setup_method(self):
        """setup any state specific to all methods of the given class"""
        pass

    def teardown_method(self):
        """teardown any state that was previously setup for all methods of the given class"""
        pass

    def test_thread_info(self):
        log_params = {
            "logger_name": "main-logger",
            "log_level": "INFO",
            "log_to_term": True,
            "log_to_file": False,
        }

        worker_log_params = {
            "logger_name": "worker-logger",
            "log_level": "INFO",
            "log_to_term": True,
            "log_to_file": False,
            "log_file": None,
        }

        client = local_client(
            cores=2,
            memory_limit="8GB",
            log_params=log_params,
            worker_log_params=worker_log_params,
            serial_execution=False,
        )

        memory_per_thread = -1
        n_threads = 0

        # Not sure if this test is deterministic. The tests are done using GitHub actions, and I am sure the container
        # environment will change over time. Hopefully, the test pulls out the most consistently calculated case.

        worker_items = client.cluster.scheduler_info["workers"].items()

        for worker_name, worker in worker_items:
            temp_memory_per_thread = (worker["memory_limit"] / worker["nthreads"]) / (
                1024**3
            )
            n_threads = n_threads + worker["nthreads"]

            if (memory_per_thread == -1) or (
                memory_per_thread > temp_memory_per_thread
            ):
                memory_per_thread = temp_memory_per_thread

        assert menrva.MenrvaClient.thread_info()["n_threads"] == n_threads
        assert (
            menrva.MenrvaClient.thread_info()["memory_per_thread"] == memory_per_thread
        )

        client.shutdown()



import logging
import multiprocessing
import os
import pathlib
import dask
import dask_jobqueue
import distributed
import psutil
import inspect
import functools

from importlib import import_module
from importlib.util import find_spec
from typing import Dict, Union

import toolviper.dask.menrva
import toolviper.utils.console as console
import toolviper.utils.logger as logger
import toolviper.utils.parameter as parameter

colorize = console.Colorize()


def load_libraries(name: str, libs: Union[str, list[str]]) -> dict[str, bool]:
    """Load libraries if they were installed and can be loaded.

    Parameters
    ----------
    name : library group name
        A library group name based on a function of a distributed environment will be imported.
    libs : Union[str, list[str]]
        a library or a list of libraries to import

    Returns
    -------
        an item of dict has the name and the flag whether all libraries were loaded successfully.
    """

    def _load_library(_lib):
        if find_spec(_lib) is not None:
            import_module(_lib)
            return [True, f"   {colorize.blue(_lib)} is available"]
        else:
            return [False, f"   {colorize.blue(_lib)} is unavailable"]

    if isinstance(libs, list):
        _tmp = list(map(_load_library, libs))
        _avail = [all([x[0] for x in _tmp]), [x[1] for x in _tmp]]
    elif isinstance(libs, str):
        _tmp = _load_library(libs)
        _avail = [_tmp[0], [_tmp[1]]]
    else:
        _avail = [False, "   illegal module specification"]

    _result = "Success" if _avail[0] else "Fail"
    logger.info(f"Loading module: {name} -- {_result}")
    [logger.info(x) for x in _avail[1]]

    return {name: _avail[0]}


def print_libraries_availability(spec: dict[str, bool]):
    """Print the contents of available_specs.

    Parameters
    ----------
    spec : dict[str, bool]
        an instance of available_specs
    """
    loaded_lib = [k for k, v in spec.items() if v]
    logger.debug(
        f"{colorize.green('Available functions of this environment')}: {', '.join(loaded_lib)}"
    )


def get_thread_info() -> Dict[str, float]:
    # This just brings the built-in thread info function into the client module.
    return toolviper.dask.menrva.MenrvaClient.thread_info()


def get_client() -> Union[None, distributed.Client]:
    """
    Get a toolviper client instance
    Returns: None or a toolviper client instance

    """
    try:
        client = distributed.Client.current()

    except ValueError:
        client = None

    if client is None:
        logger.info("There are currently no client instances.")
        return None

    return client


def get_cluster() -> Union[None, distributed.LocalCluster]:
    """
    Get a toolviper cluster instance
    Returns: None or a toolviper cluster instance

    """
    cluster = None

    if get_client() is not None:
        cluster = distributed.Client.current().cluster

    if cluster is None:
        logger.info("There are currently no cluster instances.")
        return None

    return cluster


@parameter.validate()
def local_client(
    cores: int = None,
    memory_limit: str = None,
    autorestrictor: bool = False,
    dask_local_dir: str = None,
    local_dir: str = None,
    wait_for_workers: bool = True,
    log_params: Union[None, Dict] = None,
    worker_log_params: Union[None, Dict] = None,
    dashboard_address: str = ":8787",
    serial_execution: bool = False,
) -> Union[distributed.Client, None]:
    """ Creates a local client, scheduler and workers using Dask Distributed LocalCluster (https://docs.dask.org/en/stable/deploying-python.html#reference)
    with Dask configuration tuned for VIPER and the option to use autorestrictor plugin and local cache.

    Parameters
    ----------
    cores : int
        Number of cores in Dask cluster, defaults to None
    memory_limit : str
        Amount of memory per core. It is suggested to use '8GB', defaults to None
    autorestrictor : bool
        Boolean determining usage of autorestrictor plugin, defaults to False
    dask_local_dir : str
        Where Dask should store temporary files, defaults to None. If None Dask will use \
        `./dask-worker-space`, defaults to None
    local_dir : str
        Defines client local directory, defaults to None

    wait_for_workers : bool
        Boolean determining usage of wait_for_workers option in dask, defaults to False
    log_params : dict
        The logger for the main process (code that does not run in parallel), defaults to {}
    worker_log_params : dict
        worker_log_params: Keys as same as log_params, default values given in `Additional \
        Information`_.

    dashboard_address: str
        Address on which to listen for the Bokeh diagnostics server like ‘localhost:8787’ or ‘0.0.0.0:8787’. Defaults to ‘:8787’.
        Set to None to disable the dashboard. Use ‘:0’ for a random port. See https://docs.dask.org/en/stable/deploying-python.html#reference for more information.

    serial_execution : bool
        This is an option that forces dask to run in serial mode while also setting up the logger to work. This is
        really only appropriate for debugging.

    .. _Description:

    ** _log_params **

    The log_params (worker_log_params) dictionary stores initialization information for the logger and associated
    workers. the following are the acceptable key: value pairs and their usage information.

    log_params["logger_name"] : str
        Defines the logger name to use
    log_params["log_to_term"] : bool
        Should messages log to the terminal output.
    log_params["log_level"] : str
        Defines logging level, valid options:
            - DEBUG
            - INFO
            - WARNING
            - ERROR
            - CRITICAL

        Only messages flagged as at the given level or below are logged.

    log_params["log_to_file"] : str
        Should messages log to file.

    log_params["log_file"] : str
        Name of log file to create. If none is given, the file name 'logger' will be used.

    Returns
    -------
        Dask Distributed Client
    """

    if log_params is None:
        log_params = {}

    log_params = {
        **{
            "logger_name": "client",
            "log_to_term": True,
            "log_level": "INFO",
            "log_to_file": False,
            "log_file": "client.log",
        },
        **log_params,
    }

    if worker_log_params is None:
        worker_log_params = {}

    worker_log_params = {
        **{
            "logger_name": "worker",
            "log_to_term": True,
            "log_level": "INFO",
            "log_to_file": False,
            "log_file": "client_worker.log",
        },
        **worker_log_params,
    }

    # If the user wants to change the global logger name from the
    # default value of toolviper
    os.environ["VIPER_LOGGER_NAME"] = log_params["logger_name"]

    if local_dir:
        os.environ["CLIENT_LOCAL_DIR"] = local_dir
        local_cache = True
    else:
        local_cache = False

    logger.setup_logger(**log_params)

    if dask_local_dir is None:
        logger.warning(
            f"It is recommended that the local cache directory be set using "
            f"the {colorize.blue('dask_local_dir')} parameter."
        )

    _set_up_dask(dask_local_dir)

    # This will work as long as the scheduler path isn't in some outside directory. Being that it is a plugin specific
    # to this module, I think keeping it static in the module directory it good.
    plugin_path = str(pathlib.Path(__file__).parent.resolve().joinpath("plugins/"))

    if local_cache or autorestrictor:
        dask.config.set(
            {"distributed.scheduler.preload": os.path.join(plugin_path, "scheduler.py")}
        )

        dask.config.set(
            {
                "distributed.scheduler.preload-argv": [
                    "--local_cache",
                    local_cache,
                    "--autorestrictor",
                    autorestrictor,
                ]
            }
        )

    if serial_execution:
        # Override the default behavior for debugging purposes and run synchronous
        dask.config.set(scheduler="synchronous")

        logger.info("Running client in synchronous mode.")

        return None

    # This method of assigning a worker plugin does not seem to work when using dask_jobqueue. Consequently, using \
    # client.register_worker_plugin so that the method of assigning a worker plugin is the same for local_client\
    # and slurm_cluster_client.
    # if local_cache or worker_log_params:
    #    dask.config.set({"distributed.worker.preload": os.path.join(path,'plugins/worker.py')})
    #    dask.config.set({"distributed.worker.preload-argv": ["--local_cache",local_cache,"--log_to_term",\
    #    worker_log_params['log_to_term'],"--log_to_file",worker_log_params['log_to_file'],"--log_file",\
    #    worker_log_params['log_file'],"--log_level",worker_log_params['log_level']]})

    # setup dask.distributed based multiprocessing environment
    if cores is None:
        cores = multiprocessing.cpu_count()

    if memory_limit is None:
        memory_limit = "".join(
            (str(round((psutil.virtual_memory().available / (1024**2)) / cores)), "MB")
        )

    try:
        cluster = distributed.Client.current().cluster

    except ValueError:

        cluster = distributed.LocalCluster(
            n_workers=cores,
            threads_per_worker=1,
            processes=True,
            memory_limit=memory_limit,
            # silence_logs=logging.ERROR,  # , silence_logs=logging.ERROR #,resources={ 'GPU': 2}
            dashboard_address=dashboard_address,
        )

    try:

        client = distributed.Client.current()

    except ValueError:

        client = toolviper.dask.menrva.MenrvaClient(cluster)
        client.get_versions(check=True)

    # When constructing a graph that has local cache enabled, all workers need to be up and running.
    if local_cache or wait_for_workers:
        client.wait_for_workers(n_workers=cores)

    logger.debug(f"These are the worker log parameters:\n {worker_log_params}")
    if local_cache or worker_log_params:
        client.load_plugin(
            directory=plugin_path,
            plugin="worker",
            name="worker_logger",
            local_cache=local_cache,
            log_params=worker_log_params,
        )

    logger.info("Client " + str(client))

    return client


def distributed_client(
    cluster: None,
    dask_local_dir: str = None,
    log_params: Union[None, Dict] = None,
    worker_log_params: Union[None, Dict] = None,
) -> Union[distributed.Client, None]:
    """ Setup dask cluster and logger.

    Parameters
    ----------
    cluster
    log_params : dict
        The logger for the main process (code that does not run in parallel), defaults to {}
    worker_log_params : dict
        worker_log_params: Keys as same as log_params, default values given in `Additional \
        Information`_.

    .. _Description:

    ** _log_params **

    The log_params (worker_log_params) dictionary stores initialization information for the logger and associated
    workers. the following are the acceptable key: value pairs and their usage information.

    log_params["logger_name"] : str
        Defines the logger name to use
    log_params["log_to_term"] : bool
        Should messages log to the terminal output.
    log_params["log_level"] : str
        Defines logging level, valid options:
            - DEBUG
            - INFO
            - WARNING
            - ERROR
            - CRITICAL

        Only messages flagged as at the given level or below are logged.

    log_params["log_to_file"] : str
        Should messages log to file.

    log_params["log_filee"] : str
        Name of log file to create. If none is given, the file name 'logger' will be used.

    Returns
    -------
        Dask Distributed Client
    """

    if log_params is None:
        log_params = {}

    log_params = {
        **{
            "logger_name": "client",
            "log_to_term": True,
            "log_level": "INFO",
            "log_to_file": False,
            "log_file": "client.log",
        },
        **log_params,
    }

    if worker_log_params is None:
        worker_log_params = {}

    worker_log_params = {
        **{
            "logger_name": "worker",
            "log_to_term": True,
            "log_level": "INFO",
            "log_to_file": False,
            "log_file": "client_worker.log",
        },
        **worker_log_params,
    }

    # If the user wants to change the global logger name from the
    # default value of toolviper
    os.environ["VIPER_LOGGER_NAME"] = log_params["logger_name"]

    logger.setup_logger(**log_params)

    if dask_local_dir is None:
        logger.warning(
            f"It is recommended that the local cache directory be set using "
            f"the {colorize.blue('dask_local_dir')} parameter."
        )

    _set_up_dask(dask_local_dir)

    """
    load libraries related functions of a distributed environment
    'available_specs' contains the function name and a flag that the function was loaded successfully 
    """

    logger.debug(colorize.green("Checking functions availability:"))
    available_specs = {
        **load_libraries("slurm", "dask_jobqueue"),
        **load_libraries("dask_ssh", ["asyncssh", "jupyter_server_proxy", "paramiko"]),
        **load_libraries("CUDA", "dask_cuda"),
    }

    print_libraries_availability(available_specs)

    # This will work as long as the scheduler path isn't in some outside directory. Being that it is a plugin specific
    # to this module, I think keeping it static in the module directory it good.
    plugin_path = str(pathlib.Path(__file__).parent.resolve().joinpath("plugins/"))

    client = toolviper.dask.menrva.MenrvaClient(cluster)
    client.get_versions(check=True)
    logger.info("Created client " + str(client))
    return client


def slurm_cluster_client(
    workers_per_node: int,
    cores_per_node: int,
    memory_per_node: str,
    number_of_nodes: int,
    queue: str,
    interface: str,
    python_env_dir: str,
    dask_local_dir: str,
    dask_log_dir: str,
    exclude_nodes: str = "",
    dashboard_port: int = 8787,
    local_dir: str = None,
    autorestrictor: bool = False,
    wait_for_workers: bool = True,
    log_params: Union[None, Dict] = None,
    worker_log_params: Union[None, Dict] = None,
):
    """Creates a Dask slurm_cluster_client on a multinode cluster.

        interface eth0, ib0

    Parameters
    ----------
    workers_per_node : int
        Number of workers per node ...

    cores_per_node : int
        Number of cores per node ...

    memory_per_node : str
        Memory allocation per node ...

    number_of_nodes : int
        Number of nodes ...

    queue : str
        Destination queue for each worker job. Passed to #SBATCH -p option

    interface : str
        Network interface like ‘eth0’ or ‘ib0’. This will be used both for the Dask scheduler and the Dask workers
        interface. If you need a different interface for the Dask scheduler you can pass it through the
        scheduler_options argument: interface=your_worker_interface,
        scheduler_options={'interface': your_scheduler_interface}.

    python_env_dir : str
        Python executable used to launch Dask workers. Defaults to the Python that is submitting these jobs.

    dask_local_dir : str
        Where Dask should store temporary files, defaults to None. If None Dask will use \
        `./dask-worker-space`, defaults to None

    local_dir : str
        Defines client local directory, defaults to None

    dask_log_dir : str
        Destination directory for dask log files.

    exclude_nodes : str
        Nodes to exclude.

    dashboard_port : int
        Port to use for dashboard connection.

    autorestrictor : bool
        Boolean determining usage of autorestrictor plugin, defaults to False

    wait_for_workers : bool
        Boolean determining usage of wait_for_workers option in dask, defaults to False

    log_params : dict
        Dictionary containing parameters to using for logging.

    worker_log_params : dict
        Dictionary containing parameters to using for worker logging.

    .. _Description:

    ** _log_params **

    The log_params (worker_log_params) dictionary stores initialization information for the logger and associated
    workers. the following are the acceptable key: value pairs and their usage information.

    log_params["logger_name"] : str
        Defines the logger name to use
    log_params["log_to_term"] : bool
        Should messages log to the terminal output.
    log_params["log_level"] : str
        Defines logging level, valid options:
            - DEBUG
            - INFO
            - WARNING
            - ERROR
            - CRITICAL

        Only messages flagged as at the given level or below are logged.

    log_params["log_to_file"] : str
        Should messages log to file.

    log_params["log_filee"] : str
        Name of log file to create. If none is given, the file name 'logger' will be used.

    Returns
    -------
        distributed.Client
    """

    # https://github.com/dask/dask/issues/5577

    # from distributed import Client

    if log_params is None:
        log_params = {}

    if worker_log_params is None:
        worker_log_params = {}

    if local_dir:
        os.environ["VIPER_LOCAL_DIR"] = local_dir
        local_cache = True
    else:
        local_cache = False

    # Viper logger for code that is not part of the Dask graph. The worker logger is setup in the _worker plugin.
    # from viper._utils._logger import setup_logger

    logger.setup_logger(**log_params)

    _set_up_dask(dask_local_dir)

    """
    load libraries related functions of a distributed environment
    'available_specs' contains the function name and a flag that the function was loaded successfully 
    """

    logger.debug(colorize.green("Checking functions availability:"))
    available_specs = {
        **load_libraries("slurm", "dask_jobqueue"),
        **load_libraries("dask_ssh", ["asyncssh", "jupyter_server_proxy", "paramiko"]),
        **load_libraries("CUDA", "dask_cuda"),
    }

    print_libraries_availability(available_specs)

    plugin_path = str(pathlib.Path(__file__).parent.resolve().joinpath("plugins/"))

    if local_cache or autorestrictor:
        dask.config.set(
            {"distributed.scheduler.preload": os.path.join(plugin_path, "scheduler.py")}
        )
        dask.config.set(
            {
                "distributed.scheduler.preload-argv": [
                    "--local_cache",
                    local_cache,
                    "--autorestrictor",
                    autorestrictor,
                ]
            }
        )

    # This method of assigning a worker plugin does not seem to work when using dask_jobqueue. Consequently, using
    # client.register_plugin so that the method of assigning a worker plugin is the same for local_client and
    # slurm_cluster_client.
    #
    # if local_cache or worker_log_params:
    #    dask.config.set({"distributed.worker.preload": os.path.join(plugin_path,"_utils/_worker.py")})
    #    dask.config.set({
    #    "distributed.worker.preload-argv": [
    #    "--local_cache",local_cache,
    #    "--log_to_term",worker_log_params["log_to_term"],
    #    "--log_to_file",worker_log_params["log_to_file"],
    #    "--log_file",worker_log_params["log_file"],
    #    "--log_level",worker_log_params["log_level"]]
    #    })
    #

    cluster = dask_jobqueue.SLURMCluster(
        processes=workers_per_node,
        cores=cores_per_node,
        interface=interface,
        memory=memory_per_node,
        walltime="24:00:00",
        queue=queue,
        name="viper",
        python=python_env_dir,
        local_directory=dask_local_dir,
        log_directory=dask_log_dir,
        job_extra_directives=["--exclude=" + exclude_nodes],
        # job_extra_directives=["--exclude=nmpost087,nmpost089,nmpost088"],
        scheduler_options={"dashboard_address": ":" + str(dashboard_port)},
    )  # interface="ib0"

    client = toolviper.dask.menrva.MenrvaClient(cluster)

    cluster.scale(workers_per_node * number_of_nodes)

    # When constructing a graph that has local cache enabled all workers need to be up and running.

    if local_cache or wait_for_workers:
        client.wait_for_workers(n_workers=workers_per_node * number_of_nodes)

    if local_cache or worker_log_params:
        client.load_plugin(
            directory=plugin_path,
            plugin="worker",
            name="worker_logger",
            local_cache=local_cache,
            log_params=worker_log_params,
        )

    logger.info("Created client " + str(client))

    return client


def auto_client():
    def function_wrapper(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            persistent_client = False

            if not get_client() is None:
                client = get_client()
                persistent_client = True
            else:

                # Get client inputs if they exist
                arguments = inspect.getcallargs(function, *args, **kwargs)
                if "client" in kwargs.keys():
                    client = local_client(**kwargs["client"])

                else:
                    client = local_client()

            try:
                print(f"Dask dashboard started at: {client.dashboard_link}")

                # Run the decorated function
                result = function(*args, **kwargs)

                return result
            finally:
                # Ensure the client is closed even if the function raises an exception
                if not persistent_client:
                    client.shutdown()

        return wrapper

    return function_wrapper


def _set_up_dask(local_directory):
    if local_directory:
        dask.config.set({"temporary_directory": local_directory})

    dask.config.set({"distributed.scheduler.allowed-failures": 10})
    dask.config.set({"distributed.scheduler.work-stealing": True})
    dask.config.set({"distributed.scheduler.unknown-task-duration": "99m"})
    dask.config.set({"distributed.worker.memory.pause": False})
    dask.config.set({"distributed.worker.memory.terminate": False})
    # dask.config.set({"distributed.worker.memory.recent-to-old-time": "999s"})
    dask.config.set({"distributed.comm.timeouts.connect": "3600s"})
    dask.config.set({"distributed.comm.timeouts.tcp": "3600s"})
    dask.config.set({"distributed.nanny.environ.OMP_NUM_THREADS": 1})
    dask.config.set({"distributed.nanny.environ.MKL_NUM_THREADS": 1})
    # https://docs.dask.org/en/stable/how-to/customize-initialization.html
